/*
 * $Id: inputprefs.c,v 1.21 2012/08/16 13:40:00 jacadcaps Exp $
 */
#define AROS_ALMOST_COMPATIBLE

#include "global.h"
#include <devices/keymap.h>
#include <devices/input.h>
#include <exec/resident.h>

#include <proto/charsets.h>

#include <string.h>

#define DEBUG 0
#include <aros/debug.h>

#define MESSDEBUG(x)	;
#define	DEBUG_KEYMAP(x)	;
#define	DEBUG_BOGUS_KEYMAP(x)	x;

BOOL dokey = TRUE;
BOOL domouse = TRUE;

#define FTCODEPAGE_ENV "ENV:ftcodepage"
#define FTCODEPAGE_ENVARC "ENVARC:ftcodepage"

#define CODEPAGE_ENV "ENV:CODEPAGE"
#define CODEPAGE_ENVARC "ENVARC:CODEPAGE"

#define ISO_1 "ISO-8859-1"

static
struct Resident *FindLibResident(BPTR	MySegList)
{
  struct Resident	*MyResident;
  ULONG			*MySegment;
  UWORD			*MyBuffer;
  UWORD			*EndBuffer;

  DEBUG_KEYMAP(dprintf("%s: SegList 0x%lx\n",__func__,
						 MySegList));

  while (MySegList)
  {
#define RESIDENT_SIZEOF 26
	MySegment	=(ULONG*) BADDR(MySegList);
	MyBuffer	=(UWORD*) &MySegment[1];
	EndBuffer	=(UWORD*) (((UBYTE *) MySegment) + MySegment[-1] - RESIDENT_SIZEOF - sizeof(ULONG));

	DEBUG_KEYMAP(dprintf("%s: Segment Start 0x%lx End 0x%lx\n",__func__,
						   MyBuffer,
						   EndBuffer));

	while (MyBuffer < EndBuffer)
	{
	  if (*MyBuffer == RTC_MATCHWORD)
	  {
		DEBUG_KEYMAP(dprintf("%s: Matchword found at 0x%lx\n",__func__,
							   MyBuffer));

		MyResident=(struct Resident*) MyBuffer;
		if (MyResident->rt_MatchTag == MyResident)
		{
		  DEBUG_KEYMAP(dprintf("%s: Resident 0x%lx\n",__func__,
								 MyResident));
		  return MyResident;
		}
		else
		{
		  DEBUG_KEYMAP(dprintf("%s: rt_MatchTag 0x%lx != Resident 0x%lx\n",__func__,
								 MyResident->rt_MatchTag,
								 MyResident));
		}
	  }
	  MyBuffer++;
	}
	MySegList	=(BPTR) MySegment[0];
  }

  DEBUG_KEYMAP(dprintf("%s: failed\n",__func__));

  return NULL;
}

static struct KeyMapNode *FindKeyMap(struct KeyMapResource *kmresource, CONST_STRPTR name)
{
		struct KeyMapNode *keymap = NULL, *node;

		Forbid();
		ForeachNode(&kmresource->kr_List, node)
		{
			if (!Stricmp(node->kn_Node.ln_Name, name))
			{
				keymap = node;
				break;
			}
		}
		Permit();

		return keymap;
}

static inline void CharsetNotify(void)
{
	struct Library *CharsetsBase = OpenLibrary("charsets.library", 53);
	
	if (CharsetsBase)
	{
		if ((CharsetsBase->lib_Version > 53) ||
			(CharsetsBase->lib_Revision >= 2))
		{
			SystemCharsetChanged();
		}
		
		CloseLibrary(CharsetsBase);
	}
}

void SetKeyMap(CONST_STRPTR name,BOOL doftcodepage)
{
	struct KeyMapResource *kmresource;

	kmresource = OpenResource("keymap.resource");

	if (kmresource)
	{
		struct KeyMapNode *keymap;
		BPTR seg = 0;
		BOOL found = FALSE;
		struct Resident *MyResident = NULL;
		CONST_STRPTR filepart = FilePart(name);

		/* Quick check to support built-in keymaps and to avoid needless
		 * LoadSeg/UnLoadSeg if the keymap is loaded already. Note:
		 * Keymaps can never be unloaded. - Piru
		 */
		keymap = FindKeyMap(kmresource, filepart);
		if (keymap)
		{
			found = TRUE;
		}
		else
		{
			if (filepart != name)
			{
				/* Absolute path (well only absolute paths
				 * make sense here).
				 */
				seg = LoadSeg(name);
			}
			else
			{
				UBYTE buf[128];

				strcpy(buf, "KEYMAPS:");
				AddPart(buf, name, sizeof(buf));

				DEBUG_KEYMAP(dprintf("%s: Keymap <%s>\n",__func__,buf));

				seg = LoadSeg(buf);

				if (!seg)
				{  
					strcpy(buf, "MOSSYS:Devs/Keymaps");
					AddPart(buf, name, sizeof(buf));

					DEBUG_KEYMAP(dprintf("%s: Keymap <%s>\n",__func__,buf));
					seg = LoadSeg(buf);
				}
			}
	
			DEBUG_KEYMAP(dprintf("%s: seg 0x%lx\n",__func__,seg));

			if (seg)
			{
				if ((MyResident = FindLibResident(seg)))
				{
					DEBUG_KEYMAP(dprintf("%s: Resident 0x%lx found\n",__func__,
								MyResident));
					DEBUG_KEYMAP(dprintf("%s: Name <%s>\n",__func__,
								MyResident->rt_Name));
					DEBUG_KEYMAP(dprintf("%s: ID <%s>\n",__func__,
								MyResident->rt_IdString));
					DEBUG_KEYMAP(dprintf("%s: Type %ld\n",__func__,
								MyResident->rt_Type));
					DEBUG_KEYMAP(dprintf("%s: Version %ld\n",__func__,
								MyResident->rt_Version));
					if (MyResident->rt_Flags & RTF_EXTENDED)
					{
						DEBUG_KEYMAP(dprintf("%s: Revision %ld\n",__func__,
									MyResident->rt_Revision));
					}

					keymap = (struct KeyMapNode*) MyResident->rt_Init;
				}
				else
				{
					keymap = (APTR)((LONG*)BADDR(seg) + 1);
				}
			}
		}
		DEBUG_KEYMAP(dprintf("%s: Keymap 0x%lx\n",__func__,
					keymap));

		if (keymap)
		{
			char *codepage = NULL;
			struct ExtendedKeyMapNode *ekeymap = NULL;

			DEBUG_KEYMAP(dprintf("%s: Keymap Name <%s>\n",__func__,
						keymap->kn_Node.ln_Name));

			/* Keep lock so that we don't end up adding dupe entry */
			Forbid();

			if (!found) /* only need to do this if such entry wasn't found already */
			{
				struct KeyMapNode *dupe = FindKeyMap(kmresource, filepart);
				if (dupe)
				{
					/* Here it means we ran into race,
					 * we need to use the existing node
					 * and unload the seglist. - Piru
					 */
					found = TRUE;
					keymap = dupe;
				}
			}

			/* Check for extended KeyMapNode */
			if (keymap->kn_Node.ln_Type == NT_EXTENDED &&
			    keymap->kn_Node.ln_Pri == 'E')
			{
				ekeymap = (APTR) keymap;

				DEBUG_KEYMAP(dprintf("found extended KeyMapNode %p Resident points to %p\n",ekeymap,ekeymap->ekn_Resident));

			}

			
			DEBUG_KEYMAP(dprintf("ekeymap %p MyResident %p found %d\n",ekeymap,MyResident,found));

			if (found)
			{
				/* Get struct Resident pointer from the keymap */
				MyResident = ekeymap ? ekeymap->ekn_Resident : NULL;
			}
			else
			{
				if (ekeymap)
				{
					/* Might prove useful for something */
					ekeymap->ekn_Seglist = seg;

					if(ekeymap->ekn_Resident != MyResident)
					{
						DEBUG_BOGUS_KEYMAP(dprintf("Resident in ekn_Resident doesn't match (%p), adjusting from %p\n",ekeymap->ekn_Resident,MyResident));

						ekeymap->ekn_Resident = MyResident;
					}
				}

				AddHead(&kmresource->kr_List, (struct Node *) keymap);
				seg = 0;
			}

			DEBUG_KEYMAP(dprintf("ekeymap %p MyResident %p\n",ekeymap,MyResident));

			Permit();

			SetKeyMapDefault(&keymap->kn_KeyMap);

			QueryGetResidentAttr(MyResident,&codepage,QUERYINFOATTR_LANG_CODEPAGE);

			if (codepage && *codepage)
			{
				SetIPrefs(codepage,strlen(codepage) + 1,IP_CODEPAGE);
				SetVar("CODEPAGE",codepage,-1,GVF_GLOBAL_ONLY);

				if (doftcodepage)
				{
					/* ignore when cp == iso-1 */
					if ((Strnicmp(ISO_1,codepage,10) != 0))
					{
						BPTR envftcp;
						UWORD fttable[256];
						ULONG i;

						for (i = 0; i < 256; i++)
						{
							ULONG wchar = ToUCS4(i,NULL);
							fttable[i] = wchar > 0xFFFF ? '?' : wchar;
						}

						envftcp = Open(FTCODEPAGE_ENV,MODE_NEWFILE);
						if (envftcp)
						{
							if (Write(envftcp,&fttable,sizeof (fttable)) != sizeof (fttable))
							{
								Close(envftcp);
								DeleteFile(FTCODEPAGE_ENV);
							}
							else
							{
								Close(envftcp);
							}
						}
					}
					else
					{
						// this isn't needed for iso-1
						DeleteFile(FTCODEPAGE_ENV);
					}
				}
			}
			else
			{
				/* restore the default */
				SetIPrefs(ISO_1,11,IP_CODEPAGE);
				DeleteFile(FTCODEPAGE_ENV);
				SetVar("CODEPAGE",ISO_1,-1,GVF_GLOBAL_ONLY);
			}
			
			/* make sure charsets.library knows about system config changes */
			CharsetNotify();
		}

		if (seg)
			UnLoadSeg(seg);
	}
	else
	{
		DEBUG_KEYMAP(dprintf("%s: no keymap.resource\n",__func__));
	}
}

AROS_UFH1(LONG, InputPrefs_Handler,
		  AROS_UFHA(struct IFFHandle *, iff, A2))
{
	AROS_LIBFUNC_INIT

	struct InputPrefs iprefs;
	struct Preferences prefs;

	D(bug("IPrefs: InputPrefs IFFParseBase %p iff %p\n", IFFParseBase, iff));

	ReadChunkBytes(iff, &iprefs, sizeof(iprefs));

	if (dokey || domouse) GetPrefs(&prefs, sizeof(prefs));

	if (dokey)
	{
		prefs.KeyRptDelay = iprefs.ip_KeyRptDelay;
		prefs.KeyRptSpeed = iprefs.ip_KeyRptSpeed;
	}

	if (domouse)
	{
		prefs.DoubleClick = iprefs.ip_DoubleClick;
		prefs.PointerTicks = iprefs.ip_PointerTicks;

		MESSDEBUG(dprintf("IPrefs: fallback, clickdelay set to %ld\n",prefs.DoubleClick.tv_micro));

		if (iprefs.ip_MouseAccel)
		{
			prefs.EnableCLI |= MOUSE_ACCEL;
		} else {
			prefs.EnableCLI &= ~MOUSE_ACCEL;
		}
	}

	if (dokey || domouse) SetPrefs(&prefs, sizeof(prefs), FALSE);

	MESSDEBUG(GetPrefs(&prefs, sizeof(prefs));dprintf("IPrefs: final clickdelay set to %ld\n",prefs.DoubleClick.tv_micro););

	if (dokey) SetKeyMap(iprefs.ip_Keymap,FALSE);

	return 0;

	AROS_LIBFUNC_EXIT
}

AROS_UFH1(LONG, InputPrefsExt_Handler,
		  AROS_UFHA(struct IFFHandle *, iff, A2))
{
	AROS_LIBFUNC_INIT

	struct InputPrefsExt iprefs;

	D(bug("IPrefs: InputPrefsExt IFFParseBase %p iff %p\n", IFFParseBase, iff));

	ReadChunkBytes(iff, &iprefs, sizeof(iprefs));

	if (domouse) SetIPrefs(&iprefs, sizeof(iprefs), IP_INPUTEXT);

	return 0;

	AROS_LIBFUNC_EXIT
}

AROS_UFH1(LONG, InputMousePrefs_Handler,
		  AROS_UFHA(struct IFFHandle *, iff, A2))
{
	AROS_LIBFUNC_INIT

	struct InputMousePrefs iprefs;

	D(bug("IPrefs: InputMousePrefs IFFParseBase %p iff %p\n", IFFParseBase, iff));

	ReadChunkBytes(iff, &iprefs, sizeof(iprefs));

	SetIPrefs(&iprefs, sizeof(iprefs), IP_INPUTMOUSE);

	domouse = FALSE;

	MESSDEBUG(dprintf("IPrefs: mouse prefs set, clicktime is %ld\n",iprefs.ip_DoubleClick.tv_micro));

	return 0;

	AROS_LIBFUNC_EXIT
}

AROS_UFH1(LONG, InputKeyboardPrefs_Handler,
		  AROS_UFHA(struct IFFHandle *, iff, A2))
{
	AROS_LIBFUNC_INIT

	struct InputKeyboardPrefs iprefs;

	D(bug("IPrefs: InputKeyboardPrefs IFFParseBase %p iff %p\n", IFFParseBase, iff));

	ReadChunkBytes(iff, &iprefs, sizeof(iprefs));

	SetIPrefs(&iprefs, sizeof(iprefs), IP_INPUTKEYBOARD);
	SetKeyMap(iprefs.ip_Keymap,FALSE);

	dokey = FALSE;

	return 0;

	AROS_LIBFUNC_EXIT
}

const UBYTE options[] =
	"Keyboard/K/A,"
	"RepDelayS/N/K,"
	"RepDelayM/N/K,"
	"RepSpeedS/N/K,"
	"RepSpeedM/N/K,"
	"GenerateFT2Map/S,"
	"/F";

struct Args
{
	STRPTR Keyboard;
	ULONG *RepDelayS;
	ULONG *RepDelayM;
	ULONG *RepSpeedS;
	ULONG *RepSpeedM;
	ULONG  GenerateFT2Map;
	STRPTR FutureArgs;
};

void LoadKeyboardPrefs(char *filename)
{
	struct InputKeyboardPrefs p;
	BPTR file;

	dokey = FALSE;

	strcpy(p.ip_Keymap, "usa1");
	p.ip_KeyRptDelay.tv_secs  = 0;
	p.ip_KeyRptDelay.tv_micro = 200000;
	p.ip_KeyRptSpeed.tv_secs  = 0;
	p.ip_KeyRptSpeed.tv_micro = 10000;
	p.ip_GenerateFT2Map = FALSE;

	if ((file = Open(filename,MODE_OLDFILE)))
	{
		BPTR OldFile;
		struct RDArgs rda;
		struct RDArgs *rda1;
		struct Args args;

		memset(&rda,0,sizeof(struct RDArgs));

		OldFile = Input();
		SelectInput(file);
		rda.RDA_Flags = RDAF_NOPROMPT | RDAF_STDIN;

		memset(&args,0,sizeof(args));

		if ((rda1 = ReadArgs(options, (LONG *)&args, &rda)))
		{
			if (args.Keyboard) stccpy(p.ip_Keymap,FilePart(args.Keyboard),16);
			if (args.RepDelayS) p.ip_KeyRptDelay.tv_secs = *args.RepDelayS;
			if (args.RepDelayM) p.ip_KeyRptDelay.tv_micro = *args.RepDelayM;
			if (args.RepSpeedS) p.ip_KeyRptSpeed.tv_secs = *args.RepSpeedS;
			if (args.RepSpeedM) p.ip_KeyRptSpeed.tv_micro = *args.RepSpeedM;
			if (args.GenerateFT2Map != 0) p.ip_GenerateFT2Map = TRUE;

			SetIPrefs(&p, sizeof(p), IP_INPUTKEYBOARD);
			/* Use full path, if available */
			SetKeyMap((CONST_STRPTR)((args.Keyboard) ? args.Keyboard : p.ip_Keymap), p.ip_GenerateFT2Map ? TRUE : FALSE);

			FreeArgs(rda1);
		}

		SelectInput(OldFile);
		Close(file);
	}
}

const UBYTE mouseoptions[] =
	"Pointer/K/A,"
	"DoubleClickS/N/K,"
	"DoubleClickM/N/K,"
	"PointerBlankS/N/K,"
	"PointerBlankM/N/K,"
	"PointerTicks/N/K,"
	"MouseBlankDelay/S,"
	"MouseBlankKey/S,"
	"A4Tech/S,"
	"Lefty/S";

struct mouseArgs
{
	char  *Pointer;
	ULONG *DoubleClickS;
	ULONG *DoubleClickM;
	ULONG *PointerBlankS;
	ULONG *PointerBlankM;
	ULONG *PointerTicks;
	ULONG MouseBlankDelay;
	ULONG MouseBlankKey;
	ULONG A4Tech;
	ULONG Lefty;
};

void LoadMousePrefs(char *filename)
{
	struct InputMousePrefs p;
	BPTR file;

	dokey = FALSE;

	strcpy(p.ip_PointerDir,"Default");
	p.ip_DoubleClick.tv_secs  = 0;
	p.ip_DoubleClick.tv_micro = 500000;
	p.ip_PointerBlank.tv_secs  = 30;
	p.ip_PointerBlank.tv_micro = 0;
	p.ip_MouseAccel = 0; //not used anymore!
	p.ip_MouseMode = 0;
	p.ip_PointerTicks = 2000;
	p.ip_Flags = IPMF_MOUSESCALING; //always set now!!!

	if ((file = Open(filename,MODE_OLDFILE)))
	{
		BPTR OldFile;
		struct RDArgs rda;
		struct RDArgs *rda1;
		struct mouseArgs args;

		memset(&rda,0,sizeof(struct RDArgs));

		OldFile = Input();
		SelectInput(file);
		rda.RDA_Flags = RDAF_NOPROMPT | RDAF_STDIN;

		memset(&args,0,sizeof(args));

		if ((rda1 = ReadArgs(mouseoptions, (LONG *)&args, &rda)))
		{
			if (args.Pointer) stccpy(p.ip_PointerDir,args.Pointer,16);
			if (args.DoubleClickS) p.ip_DoubleClick.tv_secs = *args.DoubleClickS;
			if (args.DoubleClickM) p.ip_DoubleClick.tv_micro = *args.DoubleClickM;
			if (args.PointerBlankS) p.ip_PointerBlank.tv_secs = *args.PointerBlankS;
			if (args.PointerBlankM) p.ip_PointerBlank.tv_micro = *args.PointerBlankM;
			if (args.PointerTicks) p.ip_PointerTicks = *args.PointerTicks;
			p.ip_Flags |= args.MouseBlankDelay ? IPMF_MOUSEBLANKDELAY : 0;
			p.ip_Flags |= args.MouseBlankKey ? IPMF_MOUSEBLANKKEY : 0;
			p.ip_Flags |= IPMF_MOUSESCALING; //always uses mouse scaling
			p.ip_MouseMode |= args.A4Tech ? IND_SETMOUSETYPE_A4TECH : 0;
			p.ip_MouseMode |= args.Lefty ? IND_SETMOUSETYPE_LEFTY : 0;

			SetIPrefs(&p, sizeof(p), IP_INPUTMOUSE);

			FreeArgs(rda1);
		}

		SelectInput(OldFile);
		Close(file);
	}
}

